# Generic Makefile for one or more executable programs that test a library
# Last edited on 2010-03-06 21:10:32 by stolfilocal

######################################################################
# To be included by the top-level Makefile in the program's directory.
# The client must define 
#
#   ${PROG} or ${PROGS}  test programs (whose main C sources are in current directory).
#   ${TEST_LIB}          the library under test (WITHOUT directory).
#   ${TEST_LIB_DIR}      directory with latest sources (and headers) of ${TEST_LIB}.
#
# The client may also define:
#   
#   ${IGNORE}         list of ".h" or ".c" files that should not be compiled or installed.
#   ${OTHER_LIBS}     list of additional libraries (WITH directories).
#   ${OTHER_I_FLAGS}  additional -I flags for C compilation and dependency checks.
#   ${OTHER_C_FLAGS}  additional flags for C compilation.
#   ${OTHER_LD_FLAGS} additional flags for linking.
#   
# All library names must include the extension (".a", ".so", etc.)
# The client must define the "all" target.
# The client may extend the "clean" target with double-colon rules.

ifeq "/${PROGS}" "/"
PROGS := ${PROG}
endif

ifneq "/${PROGS}" "/"
ifneq "/${TEST_LIB}" "/"
ifneq "/${TEST_LIB_DIR}" "/"

SHELL := /bin/sh

# ----------------------------------------------------------------------
# Package components and directories

# Location of headers of the library under test:
TEST_INC_DIR := ${TEST_LIB_DIR}

# Get list of all local ".h" files, and their derived objects:
HFILES := ${filter-out ${IGNORE}, ${wildcard *.h}}
HOFILES := ${subst .h,.ho,${HFILES}}

# Get list of all local ".c" files (including the main sources of ${PROGS}):
CFILES := ${filter-out ${IGNORE}, ${wildcard *.c}}
OFILES := ${subst .c,.o,${CFILES}}

# Get list of ".c" files of good main programs and their objects:
PROG_CFILES := ${addsuffix .c,${PROGS}}
PROG_OFILES := ${subst .c,.o,${PROG_CFILES}}

# Get list of non-main ".c" files and objects:
LIB_CFILES := ${filter-out ${PROG_CFILES} ${IGNORE}, ${CFILES}}
LIB_OFILES := ${subst .c,.o,${LIB_CFILES}}

#---------------------------------------------------------------------
# Phony "make" targets:
# Client should define the "all" target

.PHONY:: \
  default-action \
  all actions \
  build build-progs do-build-progs \
  clean clean-progs

#---------------------------------------------------------------------
# Default target: call the client's "all" target.
default-action: all

#---------------------------------------------------------------------
# "make actions" performs the actions listed in ${ACTIONS}, if any
actions:
ifeq "/${ACTIONS}" "/"
	@echo '$${ACTIONS} is empty'
else
	${MAKE} ${ACTIONS}
endif
  
#---------------------------------------------------------------------
# Directories to search for #include files:

I_FLAGS := \
  -I. \
  -I${TEST_INC_DIR} \
  ${OTHER_I_FLAGS}

#---------------------------------------------------------------------
# "make build" and "make build-progs" make sure that the ${DEPFILE} is
# up to date, and then recreate the object files from the source files
# as necessary.

build: build-progs
  
build-progs:
	${MAKE} do-build-progs

do-build-progs: ${PROGS}

LIBRARIES := \
  ${addprefix ${TEST_LIB_DIR}/, ${TEST_LIB}} \
  ${OTHER_LIBS}

CC := /usr/bin/gcc

C_FLAGS := \
  -Wall -Wundef \
  -Wpointer-arith \
  -Wmissing-prototypes \
  -ggdb \
  -fpcc-struct-return \
  -ffloat-store \
  -frounding-math
  
LD_FLAGS := \
  -ggdb

%.o: %.c
	${CC} -c ${C_FLAGS} ${OTHER_C_FLAGS} ${I_FLAGS} $*.c
  
%.ho: %.h
	${CC} -o $*.ho -c ${C_FLAGS} ${OTHER_C_FLAGS} ${I_FLAGS} -x c $*.h \
          || /bin/rm -f $*.ho
  
${PROGS}: ${PROG_OFILES} ${LIB_OFILES} ${LIBRARIES}
	@echo '# - - - - - - - - - - - - - - - - - - - - - - - - - - - - -'
	@echo building $@ ...
	-rm -f $@
	${CC} -o $@ ${LD_FLAGS} ${OTHER_LD_FLAGS} $@.o \
          ${LIB_OFILES} \
          ${LIBRARIES} -lm -lrt
	@echo '# - - - - - - - - - - - - - - - - - - - - - - - - - - - - -'
        
${TEST_LIB_DIR}/${TEST_LIB}: \
	${wildcard ${TEST_LIB_DIR}/*.h ${TEST_LIB_DIR}/*.c} \
	cd ${TEST_LIB_DIR} && ${MAKE} ${TEST_LIB}

#---------------------------------------------------------------------
# "make clean" deletes all derived files.
# It is a double-colon rule so that clients may add more actions.

clean::
	-/bin/rm -f *.o *.ho *.a core ${PROGS}

endif
endif
endif
# End of ${PROG} ${TEST_LIB} ${TEST_LIB_DIR} section.
######################################################################

